mod config;
mod meta;

use std::{
    fs::{File, OpenOptions},
    io,
    path::{Path, PathBuf},
};

use chrono::{SubsecRound, Utc};
use flate2::{write::GzEncoder, Compression};
use serde::{Deserialize, Serialize};

use super::{Backup, BackupType, Delta, State};
use crate::other;
pub use config::ManagerConfig;
pub use meta::MetaManager;

/// Manages a single backup layer consisting of one or more chains of backups.
pub struct Manager<T>
where
    T: Clone + Serialize + for<'de> Deserialize<'de> + std::fmt::Debug,
{
    backup_dir: PathBuf,
    dirs: Vec<(PathBuf, PathBuf)>,
    default_metadata: T,
    chain_len: u64,
    chains_to_keep: u64,
    frequency: chrono::Duration,
    chains: Vec<Vec<Backup<T>>>,
}

impl<T> Manager<T>
where
    T: Clone + Serialize + for<'de> Deserialize<'de> + std::fmt::Debug,
{
    const METADATA_FILE: &str = "alex.json";

    pub fn new<P: Into<PathBuf>>(
        backup_dir: P,
        dirs: Vec<(PathBuf, PathBuf)>,
        metadata: T,
        chain_len: u64,
        chains_to_keep: u64,
        frequency: chrono::Duration,
    ) -> Self {
        Self {
            backup_dir: backup_dir.into(),
            dirs,
            default_metadata: metadata,
            chain_len,
            chains_to_keep,
            frequency,
            chains: Vec::new(),
        }
    }

    /// Create a new backup, either full or incremental, depending on the state of the current
    /// chain.
    pub fn create_backup(&mut self) -> io::Result<()> {
        // We start a new chain if the current chain is complete, or if there isn't a first chain
        // yet
        if let Some(current_chain) = self.chains.last() {
            let current_chain_len: u64 = current_chain.len().try_into().unwrap();

            if current_chain_len >= self.chain_len {
                self.chains.push(Vec::new());
            }
        } else {
            self.chains.push(Vec::new());
        }

        let current_chain = self.chains.last_mut().unwrap();

        let mut backup = if !current_chain.is_empty() {
            let previous_backup = current_chain.last().unwrap();
            let previous_state = State::from(current_chain.iter().map(|b| &b.delta));

            Backup::create_from(
                previous_state,
                previous_backup.start_time,
                &self.backup_dir,
                &self.dirs,
            )?
        } else {
            Backup::create(&self.backup_dir, &self.dirs)?
        };

        backup.set_metadata(self.default_metadata.clone());

        current_chain.push(backup);

        self.save()?;

        Ok(())
    }

    /// Delete all backups associated with outdated chains, and forget those chains.
    pub fn remove_old_backups(&mut self) -> io::Result<()> {
        let chains_to_store: usize = self.chains_to_keep.try_into().unwrap();

        if chains_to_store < self.chains.len() {
            let mut remove_count: usize = self.chains.len() - chains_to_store;

            // We only count finished chains towards the list of stored chains
            let chain_len: usize = self.chain_len.try_into().unwrap();
            if self.chains.last().unwrap().len() < chain_len {
                remove_count -= 1;
            }

            for chain in self.chains.drain(..remove_count) {
                for backup in chain {
                    let path = Backup::path(&self.backup_dir, backup.start_time);
                    std::fs::remove_file(path)?;
                }
            }

            self.save()?;
        }

        Ok(())
    }

    /// Write the in-memory state to disk.
    pub fn save(&self) -> io::Result<()> {
        let json_file = File::create(self.backup_dir.join(Self::METADATA_FILE))?;
        serde_json::to_writer(json_file, &self.chains)?;

        Ok(())
    }

    /// Overwrite the in-memory state with the on-disk state.
    pub fn load(&mut self) -> io::Result<()> {
        let json_file = match File::open(self.backup_dir.join(Self::METADATA_FILE)) {
            Ok(f) => f,
            Err(e) => {
                // Don't error out if the file isn't there, it will be created when necessary
                if e.kind() == io::ErrorKind::NotFound {
                    self.chains = Vec::new();

                    return Ok(());
                } else {
                    return Err(e);
                }
            }
        };

        self.chains = serde_json::from_reader(json_file)?;

        Ok(())
    }

    /// Calculate the next time a backup should be created. If no backup has been created yet, it
    /// will return now.
    pub fn next_scheduled_time(&self) -> chrono::DateTime<Utc> {
        self.chains
            .last()
            .and_then(|last_chain| last_chain.last())
            .map(|last_backup| last_backup.start_time + self.frequency)
            .unwrap_or_else(chrono::offset::Utc::now)
    }

    /// Search for a chain containing a backup with the specified start time.
    ///
    /// # Returns
    ///
    /// A tuple (chain, index) with index being the index of the found backup in the returned
    /// chain.
    fn find(&self, start_time: chrono::DateTime<Utc>) -> Option<(&Vec<Backup<T>>, usize)> {
        for chain in &self.chains {
            if let Some(index) = chain
                .iter()
                .position(|b| b.start_time.trunc_subsecs(0) == start_time)
            {
                return Some((chain, index));
            }
        }

        None
    }

    /// Restore the backup with the given start time by restoring its chain up to and including the
    /// backup, in order.
    pub fn restore_backup(
        &self,
        start_time: chrono::DateTime<Utc>,
        dirs: &Vec<(PathBuf, PathBuf)>,
    ) -> io::Result<()> {
        self.find(start_time)
            .ok_or_else(|| other("Unknown layer."))
            .and_then(|(chain, index)| {
                for backup in chain.iter().take(index + 1) {
                    backup.restore(&self.backup_dir, dirs)?;
                }

                Ok(())
            })
    }

    /// Export the backup with the given start time as a new full archive.
    pub fn export_backup<P: AsRef<Path>>(
        &self,
        start_time: chrono::DateTime<Utc>,
        output_path: P,
    ) -> io::Result<()> {
        self.find(start_time)
            .ok_or_else(|| other("Unknown layer."))
            .and_then(|(chain, index)| {
                match chain[index].type_ {
                    // A full backup is simply copied to the output path
                    BackupType::Full => std::fs::copy(
                        Backup::path(&self.backup_dir, chain[index].start_time),
                        output_path,
                    )
                    .map(|_| ()),
                    // Incremental backups are exported one by one according to their contribution
                    BackupType::Incremental => {
                        let contributions = Delta::contributions(
                            chain.iter().take(index + 1).map(|b| &b.delta).rev(),
                        );

                        let tar_gz = OpenOptions::new()
                            .write(true)
                            .create(true)
                            .open(output_path.as_ref())?;
                        let enc = GzEncoder::new(tar_gz, Compression::default());
                        let mut ar = tar::Builder::new(enc);

                        // We only need to consider backups that have a non-empty contribution.
                        // This allows us to skip reading backups that have been completely
                        // overwritten by their successors anyways.
                        for (contribution, backup) in contributions
                            .iter()
                            .rev()
                            .zip(chain.iter().take(index + 1))
                            .filter(|(contribution, _)| !contribution.is_empty())
                        {
                            println!("{}", &backup);
                            backup.append(&self.backup_dir, contribution, &mut ar)?;
                        }

                        let mut enc = ar.into_inner()?;
                        enc.try_finish()
                    }
                }
            })
    }

    /// Get a reference to the underlying chains
    pub fn chains(&self) -> &Vec<Vec<Backup<T>>> {
        &self.chains
    }
}
